//==================================================================================================
//
// This code can be used to compute an approximation for the spectral distorting caused by different
// energy release mechanisms. The computation is fast and based on the Greens function computed for
// a fixed background cosmology.
//
//==================================================================================================
//
// Author: Jens Chluba
// first implementation: April  2013
// last modification:    August 2013
//
//==================================================================================================
// June 3rd, 2013: changed setup of decaying particle interpolation scheme 

//==================================================================================================
// Standards
//==================================================================================================
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <iomanip>
#include <cmath>
#include <limits.h>
#include <vector>

//==================================================================================================
// GSL lib
//==================================================================================================
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "gsl/gsl_multimin.h"
#include <gsl/gsl_linalg.h>

//==================================================================================================
// required libs
//==================================================================================================
#include "Greens.h"
#include "routines.h"
#include "physical_consts.h"
#include "Patterson.h"
#include "Cosmos.h"

//==================================================================================================
// namespaces
//==================================================================================================
using namespace std;

//==================================================================================================
// location of Greens-function data
//==================================================================================================
string Gfname=(string)GREENSPATH+"Gdatabase/Greens_data.dat";

double ze_G=1.001e+3, zs_G=5.0e+6*0.999;   // maximal and minimal redshift for Greens function data
double epsrel_glob=5.0e-7;

//#define add_G_term

//==================================================================================================
// change verbosity level
//==================================================================================================
int show_Green_mess=0;

void print_message(string mess)
{
    if(show_Green_mess>=0)
    {
        cout << "\n ====================================================="
        << "===================================================== " << endl;
        cout << " || " << mess << endl;
        cout << " ====================================================="
        << "===================================================== " << endl << endl;
    }
    
    return;
}

void set_verbosity(int v){ show_Green_mess=v; return; }

//==================================================================================================
//
// set cosmology object
//
//==================================================================================================
Cosmos cosmos;

double cosmos_Get_T0(){ return 2.726; }
double cosmos_Get_time(double z){ return cosmos.t(z); }

bool cosmology_is_set=0;

void set_cosmology()
{
    if(cosmology_is_set) return;
    
    //==============================================================================================
    // these are the parameters used for the computation of the Greens-function
    //==============================================================================================
    double Yp=0.24;
    double To=cosmos_Get_T0();
    double Omega0=0.26;
    double OmegaB=0.044;
    double OmegaL=0.0;         // OmegaL<=0 --> value is computed below using curvature value
    double OmegaK=0.0;
    double h100=0.71;
    double Nnu=3.04;
    double F=1.14;
   
    if(OmegaL<=0.0) OmegaL=1.0-Omega0-OmegaK-cosmos.calc_Orel(To, Nnu, h100);

    //==============================================================================================
    // for the interpolation function for Xe in Cosmos class
    //==============================================================================================
    int n_Xe=10000;
    double dens[7];
    
    //==============================================================================================
    if(show_Green_mess>1) print_message(" Initializing cosmology object ");
    
    dens[0]=Omega0+OmegaL+OmegaK;
    dens[1]=Omega0*pow(h100, 2);
    dens[2]=OmegaB*pow(h100, 2);
    dens[3]=OmegaL;
    dens[4]=F;
    //dens[5]=fDM_ann;
    dens[5]=0.0;
    dens[6]=1;
    
    //==============================================================================================
    // initializing cosmos object...
    //==============================================================================================
    cosmos.init(h100, To, Yp, dens, 1.0e+10, 1.0e-10, n_Xe, Nnu);
    
    cosmos.init_splines();
    
    if(show_Green_mess>=2) cosmos.display_cosmology();
    
    cosmology_is_set=1;
    
    return;
}


//==================================================================================================
//
// parameters for energy injection rates
//
//==================================================================================================
struct release_Data
{
    // parameters are defined as in Chluba & Sunyaev 2012
    double f_ann; int wave;      // annihilation efficiency fann --> fann (1-fnu) & s/p-wave sigma
    double A, k0, nS, nrun;      // adiabatic mode dissipation
    double Gamma_X, f_X;         // decaying particles f_X --> f_X (1-fnu)
    // special power spectra
    double DA_step, kstep, nSstep;
    double Abend, kbend, nSbend;
};

release_Data ERD={0, 0, 0, 0, 0, 0, 0, 0, 0.0, 1.0e+50, 1.0, 0.0, 1.0e+50, 1.0};


//==================================================================================================
//
// functions for energy injection rates
//
//==================================================================================================
vector<double> xdata;
vector<int> fGdata;
vector<double> Gshiftdata;
vector<double> YSZdata;
vector<double> DMsdata;
vector<double> DMpdata;
vector<int> Decdata;

int mem_Tgin=-10;
int mem_Tglast=-10;
int mem_rho=-10;

//==================================================================================================
//
// energy release from the dissipation of acoustic modes
//
//==================================================================================================
double P_zeta(double k){ return ERD.A*pow(k/ERD.k0, ERD.nS-1.0+0.5*ERD.nrun*log(k/ERD.k0)); }

double DP_zeta_step(double k){ return k>=ERD.kstep ? ERD.DA_step*pow(k/ERD.kstep, ERD.nSstep-1.0) : 0.0; }

double DP_zeta_bend(double k)
{ return k>=ERD.kbend ? P_zeta(ERD.kbend)*pow(k/ERD.kbend, ERD.nSbend-1.0)-P_zeta(k) : 0.0; }

double DP_zeta_bend_B(double k)
{ return ERD.Abend*pow(k/45.0, ERD.nSbend-1.0); }

void Set_P_step_params(double DA, double ks, double nSs)
{
    ERD.DA_step=DA;
    ERD.kstep=ks;
    ERD.nSstep=nSs;
    
    return;
}

void Set_P_bend_params(double kbend, double nSbend)
{
    ERD.kbend=kbend;
    ERD.nSbend=nSbend;    
    
    return;
}

void Set_P_bend_params_B(double Abend, double nSbend)
{
    ERD.Abend=Abend;
    ERD.nSbend=nSbend;
    
    return;
}

//==================================================================================================
double dQdlz_func_int(double logk, void *p)
{
    double *d=(double *)p;
    double kD=d[0], k=exp(logk);
    double dkDm2_dz=d[1];
    double D2=0.81;
    
    double P_back=P_zeta(k);
    
    double DP_bend_B=DP_zeta_bend_B(k);
    if(P_back<=DP_bend_B) P_back=DP_bend_B;
    
    P_back+=DP_zeta_step(k);
    P_back+=DP_zeta_bend(k);
    
    if(P_back<0) cout << k << " " << P_back << " (alert) " << endl;
    
    return 2.0*D2*P_back*2.0*k*k*dkDm2_dz*exp(-2.0*pow(k/kD, 2));
}

double dQdlz_func_dissipation(double zh)
{
    double kD=cosmos.k_D(zh);
    double epsrel=epsrel_glob, epsabs=1.0e-60;
    
    double data[2]={kD*const_Mpc, cosmos.dkm2_D_dz(zh)/const_Mpc/const_Mpc};
    void *p=(void *)data;
    
    double xmin=log(0.12), xmax=min(log(ERD.kstep), log(5.0*kD*const_Mpc));
    double r=Integrate_using_Patterson_adaptive(xmin, xmax, epsrel, epsabs, dQdlz_func_int, p);
    
    xmin=xmax, xmax=log(5.0*kD*const_Mpc);
    r+=Integrate_using_Patterson_adaptive(xmin, xmax, epsrel, epsabs, dQdlz_func_int, p);
    
    return r;
}

//==================================================================================================
double heating_function(double z)
{
    //==============================================================================================
    // dQ/dlogz == (1+z)*dQ/dz == dQ/dt / H(z) where dQ/dt is energy
    // release relative to CMB blackbody.
    //==============================================================================================
    double dQdlz=0.0;

    //==============================================================================================
    // parameters
    //==============================================================================================
    double Xp=cosmos.Xp(z);
    double fHe=cosmos.fHe();
    double ZHe=1.0-cosmos.NHeI(z)/(1.0e-30+cosmos.NHe(z));
    double NH=cosmos.NH(z);
    double t_cos=cosmos.t(z);
    double g_heat=(1.0+2.0*(Xp+fHe*ZHe)/(1.0+fHe) )/3.0;

    //==============================================================================================
    // dark matter annihilation
    //==============================================================================================
    if(ERD.wave==0) dQdlz+=ERD.f_ann*NH*pow(1.0+z, 3)*g_heat/(const_me*1.0e+3); // s-wave
    if(ERD.wave==1) dQdlz+=ERD.f_ann*NH*pow(1.0+z, 4)*g_heat/(const_me*1.0e+3); // p-wave
    
    //==============================================================================================
    // decaying particle
    //==============================================================================================
    dQdlz+=ERD.f_X*ERD.Gamma_X*NH*exp(-ERD.Gamma_X*t_cos)*g_heat/(const_me*1.0e+3);

    //==============================================================================================
    // conversion dQ/dt --> dQ/dlogz (one factor of 1+z cancels)
    //==============================================================================================
    dQdlz/=cosmos.rho_g_1_cm3(z)*cosmos.H(z);

    //==============================================================================================
    // dissipation of adiabatic modes [factor 1+z from dz --> dln z]
    //==============================================================================================
    dQdlz+=ERD.A>0.0 ? dQdlz_func_dissipation(z)*(1.0+z) : 0.0;
    
    //dQdlz=5.0e-8; // to set constant energy release rate
    
    return dQdlz;
}

//==================================================================================================
int mem_spline_heating_function=-100;
int np_spline_heat=256;

double heating_function_spline(double z){ return calc_spline_JC(log(z), mem_spline_heating_function); }

void setup_heating_function_spline(int np=300)
{
    //==============================================================================================
    // set spline for heating function
    //==============================================================================================
    vector<double> lzarr(np), harr(np);
    init_xarr(log(ze_G*0.9), log(zs_G*1.1), &lzarr[0], np, 1, 0);
    
    for(int k=0; k<np; k++) harr[k]=heating_function(exp(lzarr[k]));
    
    if(mem_spline_heating_function==-100)
        mem_spline_heating_function=calc_spline_coeffies_JC(np, &lzarr[0], &harr[0], "h-function");
    else
        update_spline_coeffies_JC(mem_spline_heating_function, np, &lzarr[0], &harr[0], "h-function");
    
    lzarr.clear();
    harr.clear();
    
    return;
}

//==================================================================================================
double f_Greens(double logzh, void *p)
{
    int *d=(int *)p;
    double z=exp(logzh)-1.0;
    
    return heating_function(z)*calc_spline_JC(logzh, fGdata[(int)d[0]])*exp(-pow(z/2.0e+6, 2.5));
}

double f_Greens_spline(double logzh, void *p)
{
    int *d=(int *)p;
    double z=exp(logzh)-1.0;
    
    return heating_function_spline(z)*calc_spline_JC(logzh, fGdata[(int)d[0]])*exp(-pow(z/2.0e+6, 2.5));
}

//==================================================================================================
void output_energy_release_history(string fname,
                                   double f_ann, int wave,
                                   double A, double k0, double nS, double nrun,
                                   double Gamma_X, double f_X)
{
    print_message(" Greens: output of energy release history ");

    //==============================================================================================
    // set parameters for calculation
    //==============================================================================================
    ERD.f_ann=f_ann;
    ERD.wave=wave;
    ERD.A=A;
    ERD.k0=k0;
    ERD.nS=nS;
    ERD.nrun=nrun;
    ERD.Gamma_X=Gamma_X;
    ERD.f_X=f_X;
    
    //==============================================================================================
    // set cosmology
    //==============================================================================================
    set_cosmology();
    // zX --> tX
    ERD.Gamma_X=1.0/cosmos.t(ERD.Gamma_X+10);

    //==============================================================================================
    // output
    //==============================================================================================
    int np=500;
    vector<double> zarr(np);
    init_xarr(ze_G, zs_G, &zarr[0], np, 1, 0);
        
    ofstream ofile(fname.c_str());
    ofile.precision(8);
    
    for(int k=0; k<np; k++) ofile << zarr[k] << " " << heating_function(zarr[k]) << endl;
    
    ofile.close();
    
    return;
}

void evaluate_energy_release_history(double *zarr, int np,
                                     double f_ann, int wave,
                                     double A, double k0, double nS, double nrun,
                                     double Gamma_X, double f_X)
{
    print_message(" Greens: evaluation energy release history at given redshifts ");
    
    //==============================================================================================
    // set parameters for calculation
    //==============================================================================================
    ERD.f_ann=f_ann;
    ERD.wave=wave;
    ERD.A=A;
    ERD.k0=k0;
    ERD.nS=nS;
    ERD.nrun=nrun;
    ERD.Gamma_X=Gamma_X;
    ERD.f_X=f_X;
    
    //==============================================================================================
    // set cosmology
    //==============================================================================================
    set_cosmology();
    // zX --> tX
    ERD.Gamma_X=1.0/cosmos.t(ERD.Gamma_X+10);
    
    //==============================================================================================
    // return results
    //==============================================================================================
    for(int k=0; k<np; k++) zarr[k]=heating_function(zarr[k]);
    
    return;
}

//==================================================================================================
// before calling this function the energy release data has to be set!
//==================================================================================================
double compute_distortion(int xi)
{
    double xmin=log(1.0+ze_G), xmax=log(1.0+zs_G);
    double epsrel=epsrel_glob, epsabs=1.0e-60;

    int data[1]={xi};
    void *p=(void *)data;
    
    return Integrate_using_Patterson_adaptive(xmin, xmax, epsrel, epsabs, f_Greens, p);
}

double compute_distortion_spline(int xi)
{
    double xmin=log(1.0+ze_G), xmax=log(1.0+zs_G);
    double epsrel=epsrel_glob, epsabs=1.0e-60;
    
    int data[1]={xi};
    void *p=(void *)data;
    
    return Integrate_using_Patterson_adaptive(xmin, xmax, epsrel, epsabs, f_Greens_spline, p);
}

//==================================================================================================
// normal y-distortion from low redshift universe
//==================================================================================================
double compute_YSZ(double x)
{
    double ex=exp(-x);
    double fac=2.0*const_h/const_cl/const_cl/pow(const_h_kb/cosmos_Get_T0(), 3)*1.0e+26/1.0e+3;
    
    return fac*pow(x, 4)*ex/(1.0-ex)/(1.0-ex)*(x*(1.0+ex)/(1.0-ex)-4.0);
}

double compute_M(double x)
{
    double ex=exp(-x);
    double fac=2.0*const_h/const_cl/const_cl/pow(const_h_kb/cosmos_Get_T0(), 3)*1.0e+26/1.0e+3;
    
    return fac*pow(x, 3)*ex/(1.0-ex)/(1.0-ex)*(x/2.19229-1.0);
}

double compute_G(double x)
{
    double ex=exp(-x);
    double fac=2.0*const_h/const_cl/const_cl/pow(const_h_kb/cosmos_Get_T0(), 3)*1.0e+26/1.0e+3;
    
    return fac*pow(x, 4)*ex/(1.0-ex)/(1.0-ex);
}

//==================================================================================================
// temperature shift from xs & Tgs (Number) to xs & Tg
//==================================================================================================
double compute_DI_Num(double xs, double Tgs, double Tg)
{
    double fac=2.0*const_h/const_cl/const_cl/pow(const_h_kb/cosmos_Get_T0(), 3)*1.0e+26/1.0e+3;
    double x=xs*Tgs/Tg;
    double ex=exp(-x), exs=exp(-xs);

    return fac*pow(xs, 3)*( exs/(1.0-exs) - ex/(1.0-ex) );
}

//==================================================================================================
//
// to initialize the Greens functions
//
//==================================================================================================
void initialize_Greens_function()
{
    //==============================================================================================
    //
    // read Greens-function
    //
    //==============================================================================================
    print_message(" Greens: reading Greens-function ");

    ifstream ifile(Gfname.c_str());
    
    //==============================================================================================
    // read header
    //==============================================================================================
    string sdum;
    char header[256];
    ifile.getline(header, 256);
    //
    ifile >> sdum;
    int nG;
    ifile >> nG;
    //
    ifile >> sdum;
    vector<double> lzhdata(nG);
    for(int jl=0; jl<nG; jl++){ ifile >> lzhdata[jl]; lzhdata[jl]=log(lzhdata[jl]+1.0); }
    //
    ifile >> sdum;
    vector<double> Tgin_data(nG);       // temperature of initial blackbody
    for(int jl=0; jl<nG; jl++) ifile >> Tgin_data[jl];
    //
    ifile >> sdum;
    vector<double> Tglast_data(nG);     // this is the temperature of last blackbody shift around z~50000!
    for(int jl=0; jl<nG; jl++) ifile >> Tglast_data[jl];
    //
    ifile >> sdum;
    vector<double> rhodata(nG);
    for(int jl=0; jl<nG; jl++) ifile >> rhodata[jl];
    ifile >> sdum;
    
    if(mem_Tgin!=-10) free_spline_JC(mem_Tgin);
    if(mem_Tglast!=-10) free_spline_JC(mem_Tglast);
    if(mem_rho!=-10) free_spline_JC(mem_rho);
    
    mem_Tgin=calc_spline_coeffies_JC(nG, &lzhdata[0], &Tgin_data[0], "temp-conv-function");
    mem_Tglast=calc_spline_coeffies_JC(nG, &lzhdata[0], &Tglast_data[0], "temp-conv-function");
    mem_rho=calc_spline_coeffies_JC(nG, &lzhdata[0], &rhodata[0], "rho-conv-function");

    //==============================================================================================
    // read Greens-function data
    //==============================================================================================
    vector<double> Gdata(nG);
    
    for(int k=0; k<(int)fGdata.size(); k++) free_spline_JC(fGdata[k]);
    
    fGdata.clear();
    Gshiftdata.clear();
    YSZdata.clear();
    xdata.clear();
    
    while(ifile)
    {
        double dum;
        ifile >> dum;
        if(!ifile) break;
        xdata.push_back(dum);
        
        for(int jl=0; jl<nG; jl++) ifile >> Gdata[jl];
        
        ifile >> dum;
        
        int mem=calc_spline_coeffies_JC(nG, &lzhdata[0], &Gdata[0], "Greens-function");
        
        fGdata.push_back(mem);
        Gshiftdata.push_back(compute_G(xdata.back()));
        YSZdata.push_back(compute_YSZ(xdata.back()));
    }
    
    ifile.close();
    
    return;
}

//==================================================================================================
//
// main call of distortion functions
//
//==================================================================================================
void output_distortion(string fname,
                       double f_ann, int wave,
                       double A, double k0, double nS, double nrun,
                       double Gamma_X, double f_X,
                       double y, double Delta)
{
    //==============================================================================================
    // set parameters for calculation
    //==============================================================================================
    ERD.f_ann=f_ann;
    ERD.wave=wave;
    ERD.A=A;
    ERD.k0=k0;
    ERD.nS=nS;
    ERD.nrun=nrun;
    ERD.Gamma_X=Gamma_X;
    ERD.f_X=f_X;

    //==============================================================================================
    // set cosmology
    //==============================================================================================
    set_cosmology();
    // zX --> tX
    ERD.Gamma_X=1.0/cosmos.t(ERD.Gamma_X+10);
    
    //==============================================================================================
    // set spline for heating function
    //==============================================================================================
    setup_heating_function_spline(np_spline_heat);
    
    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    
    print_message(" Greens: running calculation ");
    
    for(int xi=0; xi<(int)xdata.size(); xi++)
    {
        double YSZ=y*YSZdata[xi];
        double Tshift=Gshiftdata[xi]*Delta*(1.0+Delta)+Delta*Delta/2.0*YSZdata[xi];
        double primDist=compute_distortion_spline(xi);
        
        ofile << xdata[xi]/(const_h_kb/cosmos_Get_T0())/1.0e+9 << " "
              << xdata[xi] << " "
              << YSZ << " " << Tshift << " "
              << primDist << " "
              << YSZ + Tshift + primDist
              << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
//
// main call of distortion functions
//
//==================================================================================================
double f_Greens_zh(double zh, int xi)
{ return calc_spline_JC(log(1.0+zh), fGdata[xi])*exp(-pow(zh/2.0e+6, 2.5)); }

void output_Greens_function(string fname, double zs, double ze, int np)
{
    vector<double> zarr(np);
    init_xarr(zs, ze, &zarr[0], np, 1, 0);

    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    ofile.precision(8);
    
    print_message(" Greens: output of Green's function ");
    
    for(int xi=0; xi<(int)xdata.size(); xi++)
    {
        ofile << xdata[xi]/(const_h_kb/cosmos_Get_T0())/1.0e+9 << " ";
        
        for(int k=0; k<np; k++)
        {
            double zh=zarr[k];
            double G=f_Greens_zh(zh, xi), DINum=0.0;
            
#ifdef add_G_term
            double Drho=calc_spline_JC(log(1.0+zh), mem_rho);
            double Tgin=calc_spline_JC(log(1.0+zh), mem_Tgin);     // initial blackbody temperature
            double Tglast=calc_spline_JC(log(1.0+zh), mem_Tglast); // blackbody part (with respect to #) at z=50000
            DINum=compute_DI_Num(xdata[xi], Tglast, Tgin)/Drho;    // x=hnu/kTglast; add DI for temperature shift
#endif
            //===============================================================================
            // spectrum is stored with respect to x=hnu/kTglast != hnu/kT0 --> in principle
            // need to interpolate onto new grid but the difference is only ~Drho_rho
            // relative to the distortion, which can be negelected
            //===============================================================================
            ofile << (G+DINum)*1.0e-8 << " ";
        }
    
        // standard distortions
        ofile << compute_G(xdata[xi])*1e-8 << " "
              << compute_YSZ(xdata[xi])*1e-8 << " "
              << compute_M(xdata[xi])*1e-8;
        
        ofile << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
void output_Greens_function_matrix(string fname, double zs, double ze, int np)
{
    vector<double> zarr(np);
    init_xarr(zs, ze, &zarr[0], np, 1, 0);
    
    vector<double> zarrmid(np);
    vector<double> weight(np, 1.0);
    
    // use mid-point rule to compute integral
    for(int k=0; k<np-1; k++) zarrmid[k]=exp((log(zarr[k+1])+log(zarr[k]))/2.0);
    np--;

    // use trapezoidal rule to compute integral
    //for(int k=0; k<np; k++) zarrmid[k]=zarr[k];
    //weight[0]=0.5;
    //weight[np-1]=0.5;
    
    zarr=zarrmid;
    
    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    ofile.precision(16);
    
#ifdef add_G_term
    print_message(" Greens: output of Green's function matrix (with G)");
#else
    print_message(" Greens: output of Green's function matrix");
#endif
    
    ofile << 0.0 << " ";
    for(int k=0; k<np; k++) ofile << zarr[k] << " ";
    ofile << endl;
    
    for(int xi=0; xi<(int)xdata.size(); xi++)
    {
        ofile << xdata[xi]/(const_h_kb/cosmos_Get_T0())/1.0e+9 << " ";
 
        for(int k=0; k<np; k++)
        {
            double zh=zarr[k];
//            double G=f_Greens_zh(zh, xi)*exp(pow(zh/2.0e+6, 2.5));
            double G=f_Greens_zh(zh, xi);

#ifdef add_G_term
            double Drho=calc_spline_JC(log(1.0+zh), mem_rho);
            double Tgin=calc_spline_JC(log(1.0+zh), mem_Tgin);     // initial blackbody temperature
            double Tglast=calc_spline_JC(log(1.0+zh), mem_Tglast); // blackbody part (with respect to #) at z=50000
            G+=Gshiftdata[xi]*(Tglast-Tgin)/Tgin/Drho;             // x=hnu/kTglast; add DI for temperature shift
#endif
            ofile << weight[k]*G*1.0e-8 << " ";
        }

        ofile << Gshiftdata[xi]*1e-8 << " "
              << YSZdata[xi]*1e-8 << " "
              << compute_M(xdata[xi])*1e-8;
//              << f_Greens_zh(5.0e+5, xi)*1e-8*exp(pow(5.0e+5/2.0e+6, 2.5));
        
        ofile << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
//
// make mock data
//
//==================================================================================================
int selected_mem;

double Signal_function(double x, void *p){ return calc_spline_JC(x, selected_mem); }

double compute_Signal(double nuGHz, double DnuGHz) // top-hat frequency filter
{
    double x=nuGHz*1.0e+9*const_h_kb/cosmos_Get_T0();
    
    if(DnuGHz>0)
    {
        double Dx=DnuGHz*1.0e+9*const_h_kb/cosmos_Get_T0();

        double xmin=x-0.5*Dx, xmax=x+0.5*Dx;
        double epsrel=epsrel_glob, epsabs=1.0e-60;
        
        int data[1]={0};
        void *p=(void *)data;
        
        return Integrate_using_Patterson_adaptive(xmin, xmax, epsrel, epsabs, Signal_function, p)/Dx;
    }
    
    return calc_spline_JC(x, selected_mem);
}

//==================================================================================================
void compute_mock_data(double *nu_GHz, int np,
                       double f_ann, int wave,
                       double A, double k0, double nS, double nrun,
                       double Gamma_X, double f_X,
                       double y, double Delta, double Dnu_GHz, double DI_1sigma)
{
    //==============================================================================================
    // set parameters for calculation
    //==============================================================================================
    ERD.f_ann=f_ann;
    ERD.wave=wave;
    ERD.A=A;
    ERD.k0=k0;
    ERD.nS=nS;
    ERD.nrun=nrun;
    ERD.Gamma_X=Gamma_X;
    ERD.f_X=f_X;
    
    //==============================================================================================
    // set cosmology
    //==============================================================================================
    set_cosmology();
    // zX --> tX
    ERD.Gamma_X=1.0/cosmos.t(ERD.Gamma_X+10);
    
    //==============================================================================================
    // set spline for heating function
    //==============================================================================================
    setup_heating_function_spline(np_spline_heat);

    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    print_message(" Greens: running calculation of mock data ");
    
    vector<double> primDist(xdata.size());
    for(int xi=0; xi<(int)xdata.size(); xi++) primDist[xi]=compute_distortion_spline(xi);
    
    int memGt=calc_spline_coeffies_JC((int)xdata.size(), &xdata[0], &Gshiftdata[0], "G-term");
    int memYSZ=calc_spline_coeffies_JC((int)xdata.size(), &xdata[0], &YSZdata[0], "YSZ");
    int memprim=calc_spline_coeffies_JC((int)xdata.size(), &xdata[0], &primDist[0], "prim");
    
    //==============================================================================================
    //
    // compute signal over filters
    //
    //==============================================================================================
    DI_1sigma*=1.0e+26;
    
    // random number generator
    const gsl_rng_type *T;
    gsl_rng *r;
    gsl_rng_env_setup();
    //T = gsl_rng_default;
    T = gsl_rng_ranlux389;
    //T = gsl_rng_ranlxd2;
    r = gsl_rng_alloc(T);
    
    double u=0.0, dumYSZ=0.0, dumTshift=0.0;
    for(int k=0; k<np; k++)
    {
        selected_mem=memprim;
        double dumprim=compute_Signal(nu_GHz[k], Dnu_GHz);
        
        if(y!=0.0 || Delta!=0.0)
        {
            selected_mem=memYSZ;
            dumYSZ=compute_Signal(nu_GHz[k], Dnu_GHz);
        }

        if(Delta!=0.0)
        {
            selected_mem=memGt;
            double dumG=compute_Signal(nu_GHz[k], Dnu_GHz);
            dumTshift=dumG*Delta*(1.0+Delta)+dumYSZ*Delta*Delta/2.0;
        }

        if(DI_1sigma>0.0) u = gsl_ran_gaussian(r, DI_1sigma);

        nu_GHz[k]=dumprim + y*dumYSZ + dumTshift + u;
    }
    
    free_spline_JC(memGt);
    free_spline_JC(memYSZ);
    free_spline_JC(memprim);
    
    gsl_rng_free(r);

    return;
}

//==================================================================================================
void compute_mock_data(string fname,
                       double f_ann, int wave,
                       double A, double k0, double nS, double nrun,
                       double Gamma_X, double f_X,
                       double y, double Delta,
                       double nu_GHz_min, double nu_GHz_max, double Dnu_GHz, double DI_1sigma)
{
    //==============================================================================================
    // frequency grid
    //==============================================================================================
    vector<double> nuGHzvec, Sig;
    
    double nu_GHz=nu_GHz_min;
    while(nu_GHz<nu_GHz_max)
    {
        nuGHzvec.push_back(nu_GHz);
        nu_GHz+=Dnu_GHz;
    }

    Sig=nuGHzvec;
    
    //==============================================================================================
    // compute signal
    //==============================================================================================
    compute_mock_data(&Sig[0], Sig.size(),
                      f_ann, wave,
                      A, k0, nS, nrun,
                      Gamma_X, f_X,
                      y, Delta, Dnu_GHz, DI_1sigma);
    
    //==============================================================================================
    // output signal
    //==============================================================================================
    ofstream ofile(fname.c_str());
    ofile.precision(8);
    
    for(int k=0; k<(int)Sig.size(); k++)
        ofile << nuGHzvec[k] << " " << Sig[k] << " " << DI_1sigma*1.0e+26 << endl;
            
    ofile.close();
    
    return;
}


//==================================================================================================
//
// pre-binned version of the Green's function
//
//==================================================================================================
int initialize_Greens_function(double nu_GHz_min, double nu_GHz_max, double Dnu_GHz)
{
    //==============================================================================================
    //
    // read Greens-function data for all frequecies
    //
    //==============================================================================================
    print_message(" Greens: reading Greens-function (binned version) ");
    
    ifstream ifile(Gfname.c_str());
    
    //==============================================================================================
    // read header
    //==============================================================================================
    string sdum;
    char header[256];
    ifile.getline(header, 256);
    //
    ifile >> sdum;
    int nG;
    ifile >> nG;
    //
    ifile >> sdum;
    vector<double> lzhdata(nG);
    for(int jl=0; jl<nG; jl++){ ifile >> lzhdata[jl]; lzhdata[jl]=log(lzhdata[jl]+1.0); }
    //
    ifile >> sdum;
    vector<double> Tgin_data(nG);       // temperature of initial blackbody
    for(int jl=0; jl<nG; jl++) ifile >> Tgin_data[jl];
    //
    ifile >> sdum;
    vector<double> Tglast_data(nG);     // this is the temperature of last blackbody shift around z~50000!
    for(int jl=0; jl<nG; jl++) ifile >> Tglast_data[jl];
    //
    ifile >> sdum;
    vector<double> rhodata(nG);
    for(int jl=0; jl<nG; jl++) ifile >> rhodata[jl];
    ifile >> sdum;
    
    if(mem_Tgin!=-10) free_spline_JC(mem_Tgin);
    if(mem_Tglast!=-10) free_spline_JC(mem_Tglast);
    if(mem_rho!=-10) free_spline_JC(mem_rho);
    
    mem_Tgin=calc_spline_coeffies_JC(nG, &lzhdata[0], &Tgin_data[0], "temp-conv-function");
    mem_Tglast=calc_spline_coeffies_JC(nG, &lzhdata[0], &Tglast_data[0], "temp-conv-function");
    mem_rho=calc_spline_coeffies_JC(nG, &lzhdata[0], &rhodata[0], "rho-conv-function");
    
    //==============================================================================================
    // read Greens-function data
    //==============================================================================================
    vector<double> Gdata(nG), Ydata, Gtdata;
    vector<vector<double> > allGdata(nG);
    vector<double> cxdata;
    
    while(ifile)
    {
        double dum;
        ifile >> dum;
        if(!ifile) break;
        cxdata.push_back(dum);
        
        for(int jl=0; jl<nG; jl++){ ifile >> dum; allGdata[jl].push_back(dum); }
        ifile >> dum;
        
        // setup y-distortion part
        Ydata.push_back(compute_YSZ(cxdata.back()));
        // setup temperature shift part
        Gtdata.push_back(compute_G(cxdata.back()));
    }
    
    ifile.close();
    
    //==============================================================================================
    // average green's function over top-hat filters
    //==============================================================================================
    for(int k=0; k<(int)fGdata.size(); k++) free_spline_JC(fGdata[k]);
    for(int k=0; k<(int)Decdata.size(); k++) free_spline_JC(Decdata[k]);
    
    fGdata.clear();
    Gshiftdata.clear();
    YSZdata.clear();
    DMsdata.clear();
    DMpdata.clear();
    Decdata.clear();
    xdata.clear();
    
    int Ymem=calc_spline_coeffies_JC(cxdata.size(), &cxdata[0], &Ydata[0], "YSZ-function");
    int Gmem=calc_spline_coeffies_JC(cxdata.size(), &cxdata[0], &Gtdata[0], "Gt-function");

    double nu_GHz=nu_GHz_min;
    while(nu_GHz<nu_GHz_max)
    {
        for(int jl=0; jl<nG; jl++)
        {
            selected_mem=calc_spline_coeffies_JC(cxdata.size(), &cxdata[0], &allGdata[jl][0], "Greens-function");
            Gdata[jl]=compute_Signal(nu_GHz, Dnu_GHz);
            free_spline_JC(selected_mem);
        }
        
        int mem=calc_spline_coeffies_JC(nG, &lzhdata[0], &Gdata[0], "Greens-function");
        
        fGdata.push_back(mem);
        xdata.push_back(nu_GHz*1.0e+9*const_h_kb/cosmos_Get_T0());
        
        // y-distortion part
        selected_mem=Ymem;
        YSZdata.push_back(compute_Signal(nu_GHz, Dnu_GHz));

        // temperature shift part
        selected_mem=Gmem;
        Gshiftdata.push_back(compute_Signal(nu_GHz, Dnu_GHz));

        nu_GHz+=Dnu_GHz;
    }
    
    free_spline_JC(Ymem);
    free_spline_JC(Gmem);
    
    return xdata.size();
}

//==================================================================================================
double DI_norm_Decay_binned(double zh, int xi)
{ return calc_spline_JC(log(1.0+zh), Decdata[xi])*exp(-pow(zh/2.0e+6, 2.5)); }

//==================================================================================================
void compute_mock_data_binned(double *Sig, int np,
                              double f_ann, int wave,
                              double A, double k0, double nS, double nrun,
                              double Gamma_X, double f_X,
                              double y, double Delta, double DI_1sigma)
{
    //==============================================================================================
    // set parameters for calculation
    //==============================================================================================
    ERD.f_ann=ERD.A=ERD.f_X=0.0;
    ERD.wave=wave;
    ERD.k0=k0;
    ERD.nS=nS;
    ERD.nrun=nrun;

    //==============================================================================================
    // set cosmology
    //==============================================================================================
    set_cosmology();
    
    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    print_message(" Greens: running calculation of mock data (binned version) ");
    
    print_message(" Greens: setting up splines for annihilation and decaying particle scenarios ");
    
    //==============================================================================================
    // compute templates for annihilation cases
    //==============================================================================================
    if(DMsdata.size()==0 && wave==0)
    {
        ERD.f_ann=1.0;
        for(int xi=0; xi<(int)xdata.size(); xi++)
            DMsdata.push_back(compute_distortion(xi));
        
    }
    
    if(DMpdata.size()==0 && wave==1)
    {
        ERD.f_ann=1.0;
        for(int xi=0; xi<(int)xdata.size(); xi++)
            DMpdata.push_back(compute_distortion(xi));
    }

    ERD.f_ann=0.0;

    //==============================================================================================
    // compute templates for decaying particle cases
    //==============================================================================================
    if(Decdata.size()==0 && f_X>0)
    {
        ERD.f_X=1.0;
        
        int npz=256;
        vector<double> zharr(npz);
        vector<double> lzharr(npz);
        vector<double> Dec_zh(npz);
        
        init_xarr(ze_G, zs_G, &zharr[0], npz, 1, 0);
        for(int k=0; k<npz; k++) lzharr[k]=log(zharr[k]+1.0);
        
        for(int xi=0; xi<(int)xdata.size(); xi++)
        {
            Dec_zh.clear();
            
            for(int k=0; k<npz; k++)
            {
                ERD.Gamma_X=1.0/cosmos.t(zharr[k]+1.0);
                //Dec_zh.push_back( compute_distortion(xi)*exp(pow(zharr[k]/2.0e+6, 2.5)) );  // June 3rd, 2013
                Dec_zh.push_back( (zharr[k]+1.0)*compute_distortion(xi)*exp(pow(zharr[k]/2.0e+6, 2.5)) ); 
             }
            
            int mem=calc_spline_coeffies_JC(npz, &lzharr[0], &Dec_zh[0], "Decay-function");

            Decdata.push_back(mem);
        }
    }
    
    ERD.f_X=0.0;

    //==============================================================================================
    // dissipation of acoustic modes still needs to be done case by case
    // this should only be called for the dissipation effect
    //==============================================================================================
    ERD.A=A;
    if(ERD.A>0.0) setup_heating_function_spline(np_spline_heat);
    
    //==============================================================================================
    //
    // write output
    //
    //==============================================================================================
    DI_1sigma*=1.0e+26;
    
    // random number generator
    const gsl_rng_type *T;
    gsl_rng *r;
    gsl_rng_env_setup();
    //T = gsl_rng_default;
    T = gsl_rng_ranlux389;
    //T = gsl_rng_ranlxd2;
    r = gsl_rng_alloc(T);
    
    double u=0.0, prim=0.0, DM=0.0, Dec=0.0;
    
    for(int xi=0; xi<(int)xdata.size(); xi++)
    {
        if(ERD.A>0.0) prim=compute_distortion_spline(xi);
        
        if(DI_1sigma>0) u = gsl_ran_gaussian(r, DI_1sigma);
        
        if(f_ann>0.0 && wave==0) DM=f_ann*DMsdata[xi];
        if(f_ann>0.0 && wave==1) DM=f_ann*DMpdata[xi];
        if(f_X>0.0) Dec=f_X/(1.0+Gamma_X)*DI_norm_Decay_binned(Gamma_X, xi);  // June 3rd, 2013
        
        double GT=Delta*(1.0+Delta)*Gshiftdata[xi]+Delta*Delta/2.0*YSZdata[xi];
        
        Sig[xi] = prim + DM + Dec + y*YSZdata[xi] + GT + u;
    }
    
    gsl_rng_free(r);
    
    return;
}

//==================================================================================================
void compute_mock_data_binned(string fname,
                              double f_ann, int wave,
                              double A, double k0, double nS, double nrun,
                              double Gamma_X, double f_X,
                              double y, double Delta, double DI_1sigma)
{
    //==============================================================================================
    // prepare data array
    //==============================================================================================
    vector<double> Sig;
    Sig=xdata;

    compute_mock_data_binned(&Sig[0], Sig.size(),
                             f_ann, wave,
                             A, k0, nS, nrun,
                             Gamma_X, f_X,
                             y, Delta, DI_1sigma);
    
    //==============================================================================================
    //
    // write output
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    ofile.precision(8);
    
    for(int xi=0; xi<(int)xdata.size(); xi++)
    {
        ofile << xdata[xi]/(1.0e+9*const_h_kb/cosmos_Get_T0()) << " "
              << Sig[xi] << " "
              << DI_1sigma*1.0e+26 << endl;
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
//
// compute least square-fit
//
//==================================================================================================
struct minimizer_Data
{
    vector<double> xa, Ia;
    vector<double> Iapprox;
};

//--------------------------------------------------------------------------------------------------
double f_chisq(const gsl_vector *v, void *params)
{
    double Jy=gsl_vector_get(v, 0);
    double Jmu=gsl_vector_get(v, 1);
    
    minimizer_Data *p = (minimizer_Data *)params;
    
    double residual=0.0;
    for(int k=0; k<(int)p->xa.size(); k++)
    {
//        p->Iapprox[k]=Jy*0.25*compute_YSZ(p->xa[k])+(1.0-Jy)*compute_M(p->xa[k])/1.401;
        p->Iapprox[k]=Jy*0.25*compute_YSZ(p->xa[k])+Jmu*compute_M(p->xa[k])*1.401;
        
        if(p->xa[k]>=0.1 && p->xa[k]<=40) residual+=pow(p->Iapprox[k]-p->Ia[k], 2);
    }
    
    return residual;
}

//--------------------------------------------------------------------------------------------------
int find_best_fit_mu_y_Greens(vector<double> &xa, vector<double> &Ia,
                              vector<double> &solution, vector<double> &Isol)
{
    int numpar=2;
    minimizer_Data MinData;
    
    MinData.xa=xa;
    MinData.Ia=Ia;
    MinData.Iapprox.resize(Ia.size());
    
    const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex2;
    gsl_multimin_fminimizer *s = NULL;
    gsl_vector *ss, *x;
    gsl_multimin_function minex_func;
    
    size_t iter = 0;
    int status;
    double size;
    
    /* Starting point */
    x = gsl_vector_alloc (numpar);
    gsl_vector_set (x, 0, 0.5);
    gsl_vector_set (x, 1, 0.5);
    
    /* Set initial step sizes to 0.01 */
    ss = gsl_vector_alloc (numpar);
    gsl_vector_set (ss, 0, 0.0005);
    gsl_vector_set (ss, 1, 0.0005);
    
    /* Initialize method and iterate */
    minex_func.n = numpar;
    minex_func.f = f_chisq;
    minex_func.params = (void *) &MinData;
    
    s = gsl_multimin_fminimizer_alloc (T, minex_func.n);
    gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
    
    do
    {
        iter++;
        status = gsl_multimin_fminimizer_iterate(s);
        
        if(status) break;
        
        size = gsl_multimin_fminimizer_size (s);
        status = gsl_multimin_test_size (size, 1.0e-6);
        
        if (status == GSL_SUCCESS)
        {
            printf (" converged to minimum\n");
        }
    }
    while (status == GSL_CONTINUE && iter < 50000);
    
    solution.clear();
    solution.push_back(gsl_vector_get (s->x, 0));
    solution.push_back(gsl_vector_get (s->x, 1));
    Isol=MinData.Iapprox;
    
    for(int k=0; k<(int)minex_func.n; k++) cout << " " << k << " " << solution[k] << endl;
    
    gsl_vector_free(x);
    gsl_vector_free(ss);
    gsl_multimin_fminimizer_free (s);
    
    return status;
}

void output_y_visibility_function(string fname, double zs, double ze, int np)
{
    vector<double> zarr(np);
    init_xarr(zs, ze, &zarr[0], np, 1, 0);
    
    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    
    print_message(" Greens: computing best fit mu&y-distortion ");
    
    vector<double> xa, Ia(xdata.size()), solution, Isol;
    
    xa=xdata;
    
    for(int zi=0; zi<(int)zarr.size(); zi++)
    {
        double zh=zarr[zi];
        
        for(int k=0; k<(int)xdata.size(); k++) Ia[k]=f_Greens_zh(zh, k);
        
        find_best_fit_mu_y_Greens(xa, Ia, solution, Isol);
        
        ofile << zh << " " << solution[0] << " " << solution[1] << " " << 1.0-solution[0]<< endl;
         
        ofstream solf((fname+".sol_"+int_to_string(zi, 3)+".dat").c_str());
        for(int k=0; k<(int)xdata.size(); k++) solf << xa[k] << " " << Ia[k] << " " << Isol[k] << endl;
        solf.close();
    }
    
    ofile.close();
    
    return;
}

void output_best_fit_mu_y_Decay(string fname, double fX_zX, double zs, double ze, int np)
{
    vector<double> zarr(np);
    init_xarr(zs, ze, &zarr[0], np, 1, 0);
    
    //==============================================================================================
    //
    // perform calculation of spectrum according to setting
    //
    //==============================================================================================
    ofstream ofile(fname.c_str());
    
    print_message(" Greens: computing best fit mu&y-distortion for decaying particle case ");
    
    int npx=(int)xdata.size();
    vector<double> xa, Ia(npx), solution, Isol;
    
    xa=xdata;
    
    double fac=6.0e-7;
    
    for(int zi=0; zi<(int)zarr.size(); zi++)
    {
        double zh=zarr[zi];
        for(int k=0; k<npx; k++) Ia[k]=xa[k]/(1.0e+9*const_h_kb/cosmos_Get_T0());

        compute_mock_data(&Ia[0], npx,
                          0, 0,
                          0, 0, 0, 0,
                          zh, fX_zX*zh/fac,
                          0, 0, 0, 0);
        
        find_best_fit_mu_y_Greens(xa, Ia, solution, Isol);
        
        ofile << zh << " " << solution[0] << " " << solution[1] << " " << 1.0-solution[0]<< endl;
        cout << zh << " " << zi << endl;
        
        ofstream solf((fname+".sol_"+int_to_string(zi, 3)+".dat").c_str());
        for(int k=0; k<npx; k++) solf << xa[k] << " " << xa[k]/(1.0e+9*const_h_kb/cosmos_Get_T0()) << " "
                                      << Ia[k]*fac << " "
                                      << Isol[k]*fac << " "
                                      << (Ia[k]-Isol[k])*fac << endl;
        solf.close();
    }
    
    ofile.close();
    
    return;
}

//==================================================================================================
//==================================================================================================
